/*
** Module   : NameConverter.java
** Abstract : Converts Progress symbols to Java/SQL-compatible symbols.
**
** Copyright (c) 2005-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description------------------------------
** 001 ECF 20050418   @20753 Created initial version. Converts Progress
**                           symbol names to the equivalent Java variable,
**                           method, constant, or class name, or to the
**                           equivalent SQL database, table, column, or
**                           index name. Defines default substring
**                           replacements and naming conventions which can
**                           be overridden.
** 002 ECF 20050427   @20900 Better handling for multi-word replacements.
**                           Each word of the replacement is stored in a
**                           stack of pending replacements. The nextToken
**                           method returns one word at a time as a
**                           syllable. This ensures that the specialized
**                           syllable processing methods behave
**                           consistently for multi-word replacement text.
** 003 ECF 20050427   @20931 Enhancements. We now statically load a
**                           dictionary of default match phrases and their
**                           replacements. At construction, a user-defined
**                           subset of these entries are copied to the new
**                           instance. Non-standard replacements can
**                           optionally be disabled.
** 004 ECF 20050429   @20969 Fix for no-replace mode. When tokenizing but
**                           not replacing tokens, it is necessary to do a
**                           minimal sweep for illegal characters. The set
**                           of standard replacements is used for this
**                           purpose.
** 005 ECF 20050608   @21448 Changed NameConverter(String name, int type)
**                           constructor default behavior. Now replacement
**                           is enabled by default.
** 006 ECF 20050629   @21595 Made a NameConverter instance reusable. This
**                           required moving the name and type arguments
**                           from the constructors to the convert method,
**                           and making the convert method synchronized.
**                           This removes a huge performance penalty
**                           overhead in cases where the same match phrase
**                           entries are used across many conversions,
**                           since these entries can be loaded once rather
**                           than for every conversion.
** 007 ECF 20050719   @21772 Fix to nextToken method. Algorithm attempted
**                           to match candidate strings which were too
**                           long. Now candidates which are longer than
**                           the remaining symbol text are eliminated as
**                           match candidates. This avoids partial matches
**                           from appearing in the output.
** 008 GES 20051028   @23141 Added generateBeanName().
** 009 ECF 20060114   @23915 Moved generateBeanName() to StringHelper.
**                           Method is needed at runtime; this class is
**                           intended for conversion use only.
** 010 GES 20060201   @24175 Detect and resolve conflicts of Java names
**                           with reserved Java keywords.
** 011 GES 20060310   @24972 Support a forced full text replacement mode
**                           which provides a way to limit the impact of
**                           a common match string if the replacement
**                           can be forced only in the case that there is
**                           a match with the entire name.
** 012 GES 20060311   @24975 Added load/reload from a named match list.
** 013 ECF 20060331   @25331 Fix to #012 (@24975). Recursive design of
**                           convert method was not truly recursive, as
**                           each invocation was made on a new instance of
**                           the class.
** 014 ECF 20061020   @30589 Fixed the convert() method. The syllable
**                           counter must not advance for an empty
**                           syllable.
** 015 ECF 20070304   @32264 Respect camelCase as a syllable delimiter
**                           when converting a name. This takes precedence
**                           over other syllable tokenizing, since it
**                           represents an explicit intent by the name's
**                           original author.
** 016 ECF 20070305   @32267 Fix to camelCase preprocessing. Was not
**                           handling blocks of capital letters properly.
** 017 ECF 20071219   @36501 Fixed private convert() implementation. When
**                           called recursively, this method must not
**                           attempt to resolve possible Java keyword
**                           conflicts. This resolution is only done for
**                           the fully converted text, not per syllable.
** 018 GES 20090424   @41930 Import change.
** 019 SVL 20090709   @43129 Added keyword "count". Name conflicts are
**                           resolved by postpending an underscore (instead of
**                           prepending it).
** 020 GES 20091117   @44419 Moved the standard symbol replacements into this
**                           file and modified the loading mechanism to allow
**                           database-specific defaults to be loaded
**                           automatically.
** 021 CA  20130308          Convert dots in names to the "dot" string.
*/

package com.goldencode.p2j.convert;

import java.util.*;
import com.goldencode.util.*;
import com.goldencode.p2j.cfg.*;

/**
 * Converter to generate a Java or SQL symbol name from a Progress symbol
 * (variable, function, procedure, filename, or schema construct). A
 * <code>NameConverter</code> instance is unaware of the type of symbol
 * being converted, but it is aware of the type of symbol which must result
 * from the conversion. The latter is specified during construction using
 * one of this class' type constants.
 * <p>
 * Sun-like naming conventions are used to generate Java variable, method,
 * "constant", and class names:
 * <ul>
 * <li>variable and method names are "camel"-cased, with the first letter
 *     always lower case (override {@link #processVariableSyllable} and
 *     {@link #processMethodSyllable}, respectively, to modify);
 * <li>"constant" names are all upper case, with syllables separated by an
 *     an underscore (override {@link #processConstantSyllable} to modify);
 * <li>class names are "camel"-cased, with the first letter always upper
 *     case (override {@link #processClassSyllable} to modify).
 * <li>conflicts with reserved Java keywords are resolved 
 * </ul>
 * The following naming conventions are used for SQL constructs:
 * <ul>
 * <li>database names are all lower case, with syllables separated by an
 *     underscore (override {@link #processDatabaseSyllable} to modify);
 * <li>table names are all lower case, with syllables separated by an
 *     underscore (override {@link #processTableSyllable} to modify);
 * <li>column names are all lower case, with syllables separated by an
 *     underscore (override {@link #processColumnSyllable} to modify);
 * <li>index names are all lower case, with syllables separated by an
 *     underscore (override {@link #processIndexSyllable} to modify).
 * </ul>
 *
 * The converter implementation includes a generic substring replacement
 * facility to permit custom naming conventions and to handle the fact that
 * Java and SQL symbols supported a narrower character set within their
 * respective symbol names. All replacement definitions are loaded from a
 * match phrase dictionary, which is project-specific.
 * <p>
 * Replacement entries can be overridden as a whole via the {@link
 * #setReplacements} method. Individual replacements can be overridden, and
 * arbitrary replacements added, via the {@link #putReplacement} method.
 * If more than one substring key matches within a symbol name, the following
 * precedence rules apply:
 * <ul>
 * <li>a match beginning earlier in the symbol string (left to right) is
 *     always preferred over an overlapping match beginning later in the
 *     symbol string;
 * <li>given two overlapping matches beginning at the same location in the
 *     symbol string, the more specific (i.e., longer) match string is always
 *     preferred over the less specific (i.e., shorter) match string.
 * </ul>
 *
 * Multi-word replacement strings are handled such that each word is treated
 * as a separate syllable in the output string.
 * <p>
 * Except for the type-specific rules described above, case is generally
 * preserved in the converted symbol name. However, substring matches for
 * the purpose of performing replacements are case-insensitive.
 * <p>
 * TODO:  conversion to Java variable and method names should match the
 * behavior of java.bean.Introspector.decapitalize().  Currently, it does so
 * in 99.9% of cases, but it seems that certain character combinations are
 * handled differently, such as when underscore follows a first, capital
 * letter.
 * <p>
 * TODO:  if the generic replacement functionality in this class is useful
 * in other contexts, this functionality should be abstracted into a parent
 * class (e.g., StringConverter). The key methods which would be moved into
 * this parent would be:
 * <ul>
 * <li>{@link #setReplacements} - no changes
 * <li>{@link #putReplacement} - no changes
 * <li>{@link #convert} - no changes
 * <li>{@link #nextToken} - no changes
 * <li>{@link #processSyllable} - would return syllable text unmodified by
 *     default; subclasses would override.
 * <li>{@link #prepareCandidates} - no changes
 * </ul>
 * All of the instance and class variables related to conversion state would
 * also move there. In addition to a no-arg default constructor, a convenience
 * constructor would require a map of replacements as its argument.
 */
public class NameConverter
implements MatchPhraseConstants
{
   /** Used when matching substrings within symbol; indicates no match */
   private static final int NO_MATCH = 0;
   
   /** Used when matching substrings within symbol; indicates partial match */
   private static final int PARTIAL_MATCH = 1;
   
   /** Used when matching substrings within symbol; indicates full match */
   private static final int FULL_MATCH = 2;
   
   /** Default replacement mappings. */
   private static Map<String, String> mappings = null;
   
   /** Default substring replacements, indexed by substring to be replaced */
   private static MatchPhraseDictionary defaults = null;
   
   // Set up default substring replacements
   static
   {
      try
      {
         mappings = new HashMap<String, String>();
         
         // Implement the default standard mappings (this may get overridden
         // by subsequent user-defined configuration files).
         mappings.put("#", "number");
         mappings.put("%", "percent");
         mappings.put("$", "dollar");
         mappings.put("&", "and");
         mappings.put("-", "");
         mappings.put("_", "");
         mappings.put("/", "");
         mappings.put(".", "dot");
         
         defaults = MatchPhraseDictionary.load(false, mappings);
      }
      catch (Exception exc)
      {
         throw new RuntimeException("Error", exc);
      }
   }
   
   /**
    * All Java language reserved keywords or named literals and some HQL
    * keywords.
    */
   private static final Set reserved = new HashSet();
   
   static
   {
      String[] reservedNames = new String[]
      {
         "abstract",
         "boolean",
         "break",
         "byte",
         "case",
         "catch",
         "char",
         "class",
         "const",
         "continue",
         "default",
         "do",
         "double",
         "else",
         "extends",
         "final",
         "finally",
         "float",
         "for",
         "goto",
         "if",
         "implements",
         "import",
         "instanceof",
         "int",
         "interface",
         "long",
         "native",
         "new",
         "package",
         "private",
         "protected",
         "public",
         "return",
         "short",
         "static",
         "strictfp",
         "super",
         "switch",
         "synchronized",
         "this",
         "throw",
         "throws",
         "transient",
         "try",
         "void",
         "volatile",
         "while",
         "true",
         "false",
         "null",
         "count"
      };
      
      reserved.addAll(Arrays.asList(reservedNames));
   }
   
   /** Map of substring replacements, indexed by substring to be replaced */
   private Map replacements = null;
   
   /** Map of standard replacements, applied in case of no-replace mode */
   private Map standards = null;
   
   /** Map of "full text only" replacements. */
   private Map<String, String> fulltxt = null;
   
   /** Array of candidates when matching substrings for replacement */
   private String[] candidates = null;
   
   /** Array of match state for each candidate during a match operation */
   private int[] matches = null;
   
   /** Original name to be converted */
   private String name = null;
   
   /** Type of conversion being performed (using this class' constants) */
   private int type = 0;
   
   /** Current position in original name symbol */
   private int index = 0;
   
   /** A substring match detected via lookahead while processing next token */
   private String pendingMatch = null;
   
   /** A stack of pending replacement syllables detected via lookahead */
   private Stack pendingStack = new Stack();
   
   /**
    * Default constructor; <code>ALL</code> default match phrases from the
    * project-specific match phrase dictionary are loaded, and replacement
    * is enabled.
    */
   public NameConverter()
   {
      this(ALL, true);
   }
   
   /**
    * Constructor which accepts a code indicating which default matches to
    * pick up from the project's match phrase dictionary. The caller also
    * chooses whether matched tokens are replaced or left in place.
    *
    * @param   phrases
    *          A bitwise-OR'd combination of {@link MatchPhraseConstants}
    *          flags indicating which set of default match phrase entries
    *          to load from the project-specific match phrase dictionary.
    * @param   replace
    *          If <code>true</code>, any token from the match phrase entries
    *          defined by <code>phrases</code> which is found within
    *          <code>name</code> is replaced with its designated replacement
    *          string. If <code>false</code>, only tokens from the standard
    *          entries (if <code>phrases</code> includes the constant
    *          <code>STANDARD</code>) are replaced. All other tokens are
    *          recognized as separate syllables, but are not replaced.
    */
   public NameConverter(int phrases, boolean replace)
   {
      if (replace)
      {
         // Get the requested replacement entries from the default match
         // phrase dictionary.
         replacements = defaults.getEntries(phrases);
      }
      else
      {
         // We will need these to sweep tokens for minimal, standard
         // replacements. This ensures we don't end up with illegal
         // characters in the output string.
         standards = defaults.getEntries(STANDARD);
         
         // Copy only standard replacement entries (if included in phrases
         // parameter) into our local replacement map.
         if ((phrases & STANDARD) != 0)
         {
            replacements = defaults.getEntries(STANDARD);
         }
         else
         {
            replacements = new HashMap();
         }
         
         // For all non-standard entries required by the phrases parameter,
         // copy the match substrings as both the replacement key and value.
         // Any token detected will be treated as a separate syllable and
         // will be replaced with itself.
         int nonStandard = (phrases & ~STANDARD);
         Map map = defaults.getEntries(nonStandard);
         Iterator iter = map.keySet().iterator();
         while (iter.hasNext())
         {
            Object key = iter.next();
            replacements.put(key, key);
         }
      }
      
      // If space is not explicitly handled already, add a replacement for
      // it.
      if (!replacements.containsKey(" "))
      {
         replacements.put(" ", "");
      }
      
      // Load the full text list.
      fulltxt = defaults.getFullText();
   }
   
   /**
    * Load or reload the match phrase dictionary definitions from the given
    * persisted match list or the default defined for the current project. 
    * Any cached version of the dictionary is replaced if it has been read
    * previously.
    *
    * @param   list
    *          The match list to use or <code>null</code> to use the default
    *          match list.
    * @param   db
    *          <code>true</code> to add the database specific default mappings,
    *          otherwise only add the code mappings.
    *
    * @throws  ConfigurationException
    *          if there is any error with the global configuration or an
    *          error loading the dictionary from persistence. A missing
    *          match list file is not fatal and <em>does not</em> trigger
    *          this exception, but does elicit a warning message to
    *          <code>stderr</code>.
    */
   public static void load(String list, boolean db)
   throws ConfigurationException
   {
      HashMap std = new HashMap<String, String>();
      
      // Implement the default standard mappings (this may get overridden
      // by subsequent user-defined configuration files).
      std.putAll(mappings);
      
      defaults = MatchPhraseDictionary.load(list, std);
      
      // Add database-specific mappings if requested AND only if they don't
      // already exist.
      if (db)
      {
         Map<String, String> map = defaults.getFullText();
         
         if (map.get("id") == null)
            defaults.putFullText("id", "identifier");
         
         if (map.get("date") == null)
            defaults.putFullText("date", "dateValue");
      }
   }
   
   /**
    * Create a (mostly) Java-bean compatible getter or setter name for a
    * given field.  If the field is a <code>boolean</code> and a getter is
    * requested, then the prefix will be "is", otherwise getters are prefixed 
    * with "get".  All setters are prefixed with "set".  The name of the
    * field always has its first character uppercased which is slightly
    * different than how Sun defines the standard.
    *
    * @param    field
    *           The name to use as the basis, must be non-null and non-empty.
    * @param    getter
    *           If <code>true</code> the resulting bean name will be a
    *           getter.
    * @param    bool
    *           If <code>getter</code> is <code>true</code> and this is
    *           <code>true</code> then the resulting bean name will prefixed
    *           with "is" instead of "get".  Ignored for setters.
    *
    * @return   The resulting getter or setter name.
    *
    * @throws   IllegalArgumentException
    *           If the field name is <code>null</code> or is the empty
    *           string.
    */
   public static String generateBeanName(String  field,
                                         boolean getter,
                                         boolean bool)
   throws IllegalArgumentException
   {
      return StringHelper.generateBeanName(field, getter, bool);
   }
   
   /**
    * Get the name being converted.
    *
    * @return  Original name.
    */
   public String getName()
   {
      return name;
   }
   
   /**
    * Completely override the default substring replacement definitions with
    * the map specified. For each entry within the map, the key should be
    * the substring to be replaced, and the value should be the replacement
    * substring. Keys are case-insensitive; however, case is generally
    * preserved in the replacement value, though it may be overridden
    * depending upon the target symbol type.
    *
    * @param   replacements
    *          Map of substrings to match and the text which should replace
    *          them in the converted symbol name, as described above.
    */
   public void setReplacements(Map replacements)
   {
      this.replacements = replacements;
      candidates = null;
   }
   
   /**
    * Override an individual, default substring replacement definition, or
    * add an arbitrary substring replacement definition to the default
    * definitions.
    *
    * @param   key
    *          The substring to be replaced in the original symbol name.
    *          The match against this key is case-insensitive.
    * @param   value
    *          The text to be substituted in the target symbol name, wherever
    *          <code>key</code> is found in the original. Case is generally
    *          preserved in the replacement value, though it may be
    *          overridden depending upon the target symbol type.
    */
   public void putReplacement(String key, String value)
   {
      key = key.toLowerCase();
      replacements.put(key, value);
      candidates = null;
      
      // Update standards map if necessary.
      if (standards != null && standards.containsKey(key))
      {
         standards.put(key, value);
      }
   }
   
   /**
    * Perform the symbol name conversion and return the converted name.
    * This process breaks the original symbol into "syllables", though
    * these are not traditional, English-language syllables. A syllable
    * in this context is any substring within the original symbol which
    * precedes or succeeds a substring replacement, or which is itself
    * the replacement text resulting from such an operation.
    * <p>
    * For example, the symbol <code>test-name</code> is processed as three
    * syllables: <code>test</code> (precedes the replacement of
    * "<code>-</code>") and <code>name</code> (succeeds the replacement of
    * "<code>-</code>"). In addition, the empty string which replaces the
    * hyphen character between these two syllables is itself considered as
    * the second syllable for processing purposes, even though it does not
    * appear in the converted result:
    * <ul>
    * <li>for type TYPE_VARIABLE: <code>testName</code>
    * <li>for type TYPE_METHOD: <code>testName</code>
    * <li>for type TYPE_CONSTANT: <code>TEST_NAME</code>
    * <li>for type TYPE_CLASS: <code>TestName</code>
    * <li>for type TYPE_DATABASE: <code>test_name</code>
    * <li>for type TYPE_TABLE: <code>test_name</code>
    * <li>for type TYPE_COLUMN: <code>test_name</code>
    * <li>for type TYPE_INDEX: <code>test_name</code>
    * </ul>
    *
    * If the starting symbol were instead <code>test%name</code>, the second
    * syllable would instead be <code>Pct</code> (using the default set of
    * replacements).
    * <p>
    * The {@link #processSyllable} method is called for each syllable
    * encountered. The string returned from that method is appended to the
    * target symbol. This occurs until the entire, original symbol name has
    * been processed.
    *
    * @param   name
    *          Name to be converted.
    * @param   type
    *          One of the type constants defined in this class. Indicates
    *          what type of Java or SQL symbol should result from the
    *          conversion.
    *
    * @return  The fully converted symbol name.
    */
   public synchronized String convert(String name, int type)
   {
      return convert(name, type, false);
   }
   
   /**
    * Retrieve a string describing the conversion type performed by this
    * converter. This is useful for reporting and debugging.
    *
    * @return  Description of conversion type.
    */
   public String describeConversion()
   {
      String desc = null;
      
      switch (type)
      {
         case TYPE_SIMPLE:
            desc = "Simple";
            break;
         case TYPE_VARIABLE:
            desc = "Java variable";
            break;
         case TYPE_METHOD:
            desc = "Java method";
            break;
         case TYPE_CONSTANT:
            desc = "Java \"constant\"";
            break;
         case TYPE_CLASS:
            desc = "Java class";
            break;
         case TYPE_DATABASE:
            desc = "SQL database";
            break;
         case TYPE_TABLE:
            desc = "SQL table";
            break;
         case TYPE_COLUMN:
            desc = "SQL column";
            break;
         case TYPE_INDEX:
            desc = "SQL index";
            break;
      }
      
      return desc;
   }
   
   /**
    * Get a string representation of this object. Reports the text being
    * converted and the type of conversion performed.
    *
    * @return  String describing this object.
    */
   public String toString()
   {
      StringBuilder buf = new StringBuilder(name);
      buf.append(" (")
         .append(describeConversion())
         .append(")");
      
      return buf.toString();
   }
   
   /**
    * Process each syllable which will become a component of the converted
    * symbol name. This method is called <em>after</em> all substring
    * replacements have already been made. Depending upon the type of the
    * target symbol, this method performs different conversions on the
    * syllable text, as discussed in this class' {@link NameConverter
    * description}.
    *
    * @param   text
    *          Text which comprises the current syllable being processed.
    * @param   syllable
    *          Zero-based index of this syllable in the converted symbol
    *          name.
    *
    * @return  The current symbol after any type-specific transformation
    *          has been applied.
    */
   protected String processSyllable(String text, int syllable)
   {
      switch (type)
      {
         case TYPE_SIMPLE:
            return text;
         case TYPE_VARIABLE:
            return processVariableSyllable(text, syllable);
         case TYPE_METHOD:
            return processMethodSyllable(text, syllable);
         case TYPE_CONSTANT:
            return processConstantSyllable(text, syllable);
         case TYPE_CLASS:
            return processClassSyllable(text, syllable);
         case TYPE_DATABASE:
            return processDatabaseSyllable(text, syllable);
         case TYPE_TABLE:
            return processTableSyllable(text, syllable);
         case TYPE_COLUMN:
            return processColumnSyllable(text, syllable);
         case TYPE_INDEX:
            return processIndexSyllable(text, syllable);
         default:
            throw new IllegalArgumentException("Illegal type: " + type);
      }
   }
   
   /**
    * Process a syllable for a Java variable symbol name.
    * <p>
    * Default processing is to lowercase the first letter of the first
    * symbol, and uppercase the first letter of all subsequent syllables.
    *
    * @param   text
    *          Text of current syllable.
    * @param   syllable
    *          Zero-based index of current syllable.
    */
   protected String processVariableSyllable(String text, int syllable)
   {
      return processVarMeth(text, syllable);
   }
   
   /**
    * Process a syllable for a Java method symbol name.
    * <p>
    * Default processing is to lowercase the first letter of the first
    * symbol, and uppercase the first letter of all subsequent syllables.
    *
    * @param   text
    *          Text of current syllable.
    * @param   syllable
    *          Zero-based index of current syllable.
    */
   protected String processMethodSyllable(String text, int syllable)
   {
      return processVarMeth(text, syllable);
   }
   
   /**
    * Process a syllable for a Java "constant" symbol name.
    * <p>
    * Default processing is to uppercase all letters and prepend an
    * underscore character to each syllable after the first.
    *
    * @param   text
    *          Text of current syllable.
    * @param   syllable
    *          Zero-based index of current syllable.
    */
   protected String processConstantSyllable(String text, int syllable)
   {
      if (text.length() == 0)
      {
         return text;
      }
      
      StringBuilder buf = new StringBuilder();
      
      // Set all characters to upper case;  separate syllables with an
      // underscore.
      if (syllable > 0)
      {
         buf.append('_');
      }
      
      buf.append(text.toUpperCase());
      
      return buf.toString();
   }
   
   /**
    * Process a syllable for a Java class or interface symbol name.
    * <p>
    * Default processing is to uppercase the first letter of every syllable.
    *
    * @param   text
    *          Text of current syllable.
    * @param   syllable
    *          Zero-based index of current syllable.
    */
   protected String processClassSyllable(String text, int syllable)
   {
      if (text.length() <= 1)
      {
         return text.toUpperCase();
      }
      
      StringBuilder buf = new StringBuilder();
      
      // Set first character of every syllable to upper case.
      buf.append(Character.toUpperCase(text.charAt(0)));
      buf.append(text.substring(1));
         
      return buf.toString();
   }
   
   /**
    * Process a syllable for an SQL database symbol name.
    * <p>
    * Default processing is to lowercase all letters and prepend an
    * underscore character to each syllable after the first.
    *
    * @param   text
    *          Text of current syllable.
    * @param   syllable
    *          Zero-based index of current syllable.
    */
   protected String processDatabaseSyllable(String text, int syllable)
   {
      return processSql(text, syllable);
   }
   
   /**
    * Process a syllable for an SQL table symbol name.
    * <p>
    * Default processing is to lowercase all letters and prepend an
    * underscore character to each syllable after the first.
    *
    * @param   text
    *          Text of current syllable.
    * @param   syllable
    *          Zero-based index of current syllable.
    */
   protected String processTableSyllable(String text, int syllable)
   {
      return processSql(text, syllable);
   }
   
   /**
    * Process a syllable for an SQL column symbol name.
    * <p>
    * Default processing is to lowercase all letters and prepend an
    * underscore character to each syllable after the first.
    *
    * @param   text
    *          Text of current syllable.
    * @param   syllable
    *          Zero-based index of current syllable.
    */
   protected String processColumnSyllable(String text, int syllable)
   {
      return processSql(text, syllable);
   }
   
   /**
    * Process a syllable for an SQL index symbol name.
    * <p>
    * Default processing is to lowercase all letters and prepend an
    * underscore character to each syllable after the first.
    *
    * @param   text
    *          Text of current syllable.
    * @param   syllable
    *          Zero-based index of current syllable.
    */
   protected String processIndexSyllable(String text, int syllable)
   {
      return processSql(text, syllable);
   }
   
   /**
    * Perform the symbol name conversion and return the converted name.
    * This process breaks the original symbol into "syllables", though
    * these are not traditional, English-language syllables. A syllable
    * in this context is any substring within the original symbol which
    * precedes or succeeds a substring replacement, or which is itself
    * the replacement text resulting from such an operation.
    * <p>
    * For example, the symbol <code>test-name</code> is processed as three
    * syllables: <code>test</code> (precedes the replacement of
    * "<code>-</code>") and <code>name</code> (succeeds the replacement of
    * "<code>-</code>"). In addition, the empty string which replaces the
    * hyphen character between these two syllables is itself considered as
    * the second syllable for processing purposes, even though it does not
    * appear in the converted result:
    * <ul>
    * <li>for type TYPE_VARIABLE: <code>testName</code>
    * <li>for type TYPE_METHOD: <code>testName</code>
    * <li>for type TYPE_CONSTANT: <code>TEST_NAME</code>
    * <li>for type TYPE_CLASS: <code>TestName</code>
    * <li>for type TYPE_DATABASE: <code>test_name</code>
    * <li>for type TYPE_TABLE: <code>test_name</code>
    * <li>for type TYPE_COLUMN: <code>test_name</code>
    * <li>for type TYPE_INDEX: <code>test_name</code>
    * </ul>
    *
    * If the starting symbol were instead <code>test%name</code>, the second
    * syllable would instead be <code>Pct</code> (using the default set of
    * replacements).
    * <p>
    * The {@link #processSyllable} method is called for each syllable
    * encountered. The string returned from that method is appended to the
    * target symbol. This occurs until the entire, original symbol name has
    * been processed.
    * <p>
    * This method may be invoked as an internal attempt to remove any illegal
    * characters from a converted syllable.  In this case, full text only
    * mode is ignored.
    *
    * @param   name
    *          Name to be converted.
    * @param   type
    *          One of the type constants defined in this class. Indicates
    *          what type of Java or SQL symbol should result from the
    *          conversion.
    * @param   internal
    *          <code>true</code> if this method is invoked due to internal
    *          housekeeping;  <code>false</code> if invoked as the result of
    *          an external call.
    *
    * @return  The fully converted symbol name.
    */
   private String convert(String name, int type, boolean internal)
   {
      this.name = preprocess(name);
      this.type = type;
      index = 0;
      
      // Honor full text only mode, if specified for the given name AND not
      // an internal invocation.
      if (!internal)
      {
         String result = fullTextReplacement(this.name); 
         if (result != null)
         {
            return processSyllable(result, 0);
         }
      }
      
      prepareCandidates();
      
      StringBuilder buf = new StringBuilder();
      
      String tok = null;
      for (int i = 0; (tok = nextToken()) != null; i++)
      {
         if (standards != null)
         {
            // Sweep tok to at least apply standard replacements, in order
            // to prevent illegal characters in the output.
            NameConverter nc = new NameConverter(0, true);
            nc.setReplacements(standards);
            tok = nc.convert(tok, type, true);
         }
         
         buf.append(processSyllable(tok, i));
         
         // Adjust the syllable count downward to skip this syllable if,
         // after conversion, it's empty.
         if (tok == null || tok.length() == 0)
         {
            i--;
         }
      }
      
      String result = buf.toString();
      
      // Resolve possible keyword conflicts on the converted output;  we don't
      // do this when called recursively, because we don't want to process
      // individual syllables, just the final result.
      if (!internal)
      {
         result = resolvePossibleKeywordConflict(result, type);
      }
      
      return result;
   }
   
   /**
    * Preprocess the given name.  This involves splitting the name into
    * syllables which are already well established by camel casing, then
    * lowercasing the entire name.  Syllables are delimited by spaces in the
    * preprocessed string.
    *
    * @param   name
    *          Name to be converted.
    *
    * @return  Preprocessed name, lowercased and split into syllables divined
    *          from camel case in the original name.
    */
   private String preprocess(String name)
   {
      StringBuilder buf = new StringBuilder();
      int len = name.length();
      char last = 0;
      for (int i = 0; i < len; i++)
      {
         char c = name.charAt(i);
         if (Character.isUpperCase(c) &&
             i > 0                    &&
             !Character.isUpperCase(last))
         {
            buf.append(' ');
         }
         buf.append(Character.toLowerCase(c));
         last = c;
      }
      
      return buf.toString().trim();
   }
   
   /**
    * If the given candidate name is marked as a forced full text match (a 
    * match that is only allowed with the full text of the name rather than
    * as a partial basis), return that replacement.
    *
    * @param   candidate
    *          The string that is to be tested.
    *
    * @return  The full text replacement or <code>null</code> if this 
    *          candidate is not forced into full text mode.
    */
   private String fullTextReplacement(String candidate)
   {
      return fulltxt.get(candidate);
   }
   
   /**
    * This is the backing method behind the {@link #processVariableSyllable}
    * and {@link #processMethodSyllable} methods, which both have the same
    * default behavior.
    *
    * @param   text
    *          Text of current syllable.
    * @param   syllable
    *          Zero-based index of current syllable.
    */
   private String processVarMeth(String text, int syllable)
   {
      int len = text.length();
      if (len < 1)
      {
         return text;
      }
      
      StringBuilder buf = new StringBuilder();
      
      // Set first character of first syllable to lower case;  first
      // character of all other syllables to upper case.
      if (syllable == 0)
      {
         buf.append(Character.toLowerCase(text.charAt(0)));
      }
      else
      {
         buf.append(Character.toUpperCase(text.charAt(0)));
      }
      
      if (len > 1)
      {
         buf.append(text.substring(1));
      }
         
      return buf.toString();
   }
   
   /**
    * This is the backing method behind the {@link #processDatabaseSyllable},
    * {@link #processTableSyllable}, {@link #processColumnSyllable}, and
    * {@link #processIndexSyllable} methods, which all have the same default
    * behavior.
    *
    * @param   text
    *          Text of current syllable.
    * @param   syllable
    *          Zero-based index of current syllable.
    */
   private String processSql(String text, int syllable)
   {
      if (text.length() == 0)
      {
         return text;
      }
      
      StringBuilder buf = new StringBuilder();
      
      // Set all characters to lower case;  separate syllables with an
      // underscore.
      if (syllable > 0)
      {
         buf.append('_');
      }
      
      buf.append(text.toLowerCase());
      
      return buf.toString();
   }
   
   /**
    * Generate the next token (aka syllable) to be appended to the target
    * symbol name. A token generated by this method will either be:
    * <ul>
    * <li>a run of characters for which no substring replacement could be
    *     matched; or
    * <li>the replacement text for a run of characters which was matched.
    * </ul>
    * Each token consists only of those characters up to the next run as
    * described above, or to the end of the original symbol string for the
    * last such run.
    * <p>
    * For example, the symbol <code>this&that</code> would generate three
    * tokens (i.e., syllables), using the default replacement map:
    * <ol>
    * <li><code>this</code> - not matched for any substring replacement; its
    *     end is determined by the fact that the next character begins a
    *     a run of characters (in this case, one character, actually) which
    *     is matched to replacement key <code>&amp;</code>;
    * <li><code>and</code> - replacement text for substring
    *     <code>&amp;</code>;
    * <li><code>that</code> - not matched for any substring replacement; its
    *     end is determined by the end of the original token string.
    * </ol>
    *
    * @return  The next token, as described above, or <code>null</code> if
    *          we have reached the end of the original symbol string and
    *          we have no new characters to return.
    */
   private String nextToken()
   {
      // If we found a full match after finding one or more non-matching
      // characters in a previous pass through this method, we will have
      // a pending match. Return the next syllable of its replacement now
      // as the next token, advancing the index to just beyond the matched
      // text.
      String replacement = getPendingReplacement();
      if (replacement != null)
      {
         // Consume the pending match.
         if (pendingMatch != null)
         {
            index += pendingMatch.length();
            pendingMatch = null;
         }
         
         return replacement;
      }
      
      resetMatchStatus();
      
      // Create buffer for next token.
      StringBuilder buf = new StringBuilder();
      
      // Offset from beginning of match keys.
      int off = 0;
      
      // Number of full key matches at the current character from the
      // original symbol name.
      int full = 0;
      
      // Number of partial key matches at the current character from the
      // original symbol name.
      int partial = 0;
      
      // Index of the character on which we last started a match, in case
      // the match turns out to be a bust and we have to push back lookahead
      // characters.
      int lastMatchStart = -1;
      
      // Scan a character at a time of the original symbol name, starting
      // at the current index, as advanced by previous calls to this method.
      int len = name.length();
      int i;
      for (i = 0; index + i < len; i++)
      {
         // Next character from original symbol.
         char ch = Character.toLowerCase(name.charAt(index + i));
         
         // Number of remaining characters in original symbol yet to be
         // tested.
         int remaining = len - index - i;
         
         // Walk the list of match statuses for each key and update, based
         // on whether the current character matches or not.
         for (int j = 0; j < matches.length; j++)
         {
            String key = candidates[j];
            if (matches[j] != FULL_MATCH)
            {
               int keyLen = key.length();
               if (keyLen - off <= remaining &&
                   off < keyLen &&
                   Character.toLowerCase(key.charAt(off)) == ch)
               {
                  if (off + 1 == keyLen)
                  {
                     // We have successfully matched the final character of
                     // the key. Update the key's status to a full match.
                     boolean fromPartial = (matches[j] == PARTIAL_MATCH);
                     if (fromPartial)
                     {
                        partial--;
                     }
                     
                     // Only consider this a full match if we were previously
                     // a partial match or if the key is only one character
                     // long.
                     if (fromPartial || keyLen == 1)
                     {
                        matches[j] = FULL_MATCH;
                        full++;
                     }
                  }
                  else if (matches[j] == NO_MATCH && off == 0)
                  {
                     // We have successfully matched the first character
                     // Update the key's status to a partial match.
                     partial++;
                     matches[j] = PARTIAL_MATCH;
                  }
               }
               else if (matches[j] == PARTIAL_MATCH)
               {
                  // We had a partial match, but the current charater did
                  // not match. Update the key's status to no match.
                  matches[j] = NO_MATCH;
                  partial--;
               }
            }
         }
         
         /*
         // DEBUG CODE
         System.out.println("ch = " + ch);
         System.out.println("candidates = " + Arrays.asList(candidates));
         System.out.print("matches = [");
         for (int x = 0; x < matches.length; x++)
         {
            if (x > 0)
               System.out.print(", ");
            System.out.print(String.valueOf(matches[x]));
         }
         System.out.println("]");
         System.out.println("i = " + i);
         System.out.println("index = " + index);
         System.out.println("len = " + len);
         System.out.println("partial = " + partial);
         System.out.println("full = " + full);
         System.out.println();
         */
         
         if (partial > 0 || full > 0)
         {
            // Increment the key offset if any key had any match.
            off++;
            
            // Remember where this match began, in case it fails and we
            // need to push back the lookahead.
            if (lastMatchStart < 0)
            {
               lastMatchStart = i;
            }
         }
         else if (partial == 0 && full == 0)
         {
            // Reset the key offset if no keys had a match.
            off = 0;
            
            // Conditionally push back.
            if (lastMatchStart >= 0)
            {
               resetMatchStatus();
               
               // Push back all characters to the position just after where
               // we last matched.
               i = lastMatchStart;
               lastMatchStart = -1;
               
               // Truncate the token buffer accordingly.
               buf.setLength(i + 1);
               
               // Skip the remainder of the outer loop, because we don't
               // want to add the current character to the buffer after
               // we've truncated it.
               continue;
            }
         }
         
         if (full > 0 && (partial == 0 || (index + i) == (len - 1)))
         {
            // At least one full match has been found, AND all partial
            // matches have resolved to either a full match or no match
            // at all OR we are at the end of the input string.
            
            // Always prefer the most specific (i.e., longest), matching
            // key.
            String key = null;
            for (int k = 0; k < matches.length; k++)
            {
               if (matches[k] == FULL_MATCH)
               {
                  String next = candidates[k];
                  if (key == null || next.length() > key.length())
                  {
                     key = next;
                  }
               }
            }
            
            // Set the pending replacement word(s) into the pending stack.
            // These will be popped as replacements are returned from this
            // method.
            setupPendingStack(key);
            
            int bufLen = buf.length();
            if (lastMatchStart > 0)
            {
               // If we already found one or more non-matching characters
               // before we detected this full match, we have to return the
               // non-matching character(s) as the next token. However, we
               // set the full match into our pendingMatch instance variable,
               // so that we use it immediately on our next call into this
               // method.
               pendingMatch = key;
               bufLen = lastMatchStart;
               index += bufLen;
               buf.setLength(bufLen);
               
               return buf.toString();
            }
            else
            {
               // Our matching key was not preceded by any non-matching
               // characters, so we can return the replacement text
               // immediately.
               index += key.length();
               
               return getPendingReplacement();
            }
         }
         
         // Add the character to the current token buffer.
         buf.append(ch);
      }
      
      // Update the index to reflect all "lookahead" characters processed
      // with no match.
      index += i;
      
      // Get the string of unmatched characters.
      String token = buf.toString();
      
      // Return the buffer or null if no new characters are available.
      return (token.length() > 0 ? token : null);
   }
   
   /**
    * Reset the list of potential matches so that each candidate has an
    * equal starting state of NO_MATCH. As soon any character in a
    * candidate is found to match, its corresponding entry is set to
    * PARTIAL_MATCH. If all characters match, it is set to FULL_MATCH.
    */
   private void resetMatchStatus()
   {
      for (int i = 0; i < matches.length; i++)
      {
         matches[i] = NO_MATCH;
      }
   }
   
   /**
    * Set up the stack of pending replacement words. Each word will be
    * returned as a separate token (i.e., syllable) from {@link #nextToken}
    * until the stack is empty.
    *
    * @param   key
    *          A substring which was matched in the original symbol; the
    *          key to a replacement string in the map of replacements.
    */
   private void setupPendingStack(String key)
   {
      String replacement = (String) replacements.get(key);
      String[] words = replacement.split(" ");
      for (int i = words.length - 1; i >= 0; i--)
      {
         pendingStack.push(words[i]);
      }
   }
   
   /**
    * Get the next pending replacement string, if any, based on a previous
    * substring match.
    *
    * @return  Next pending replacement syllable, or <code>null</code> if
    *          the stack of pending replacements is empty.
    */
   private String getPendingReplacement()
   {
      String replacement = null;
      if (!pendingStack.isEmpty())
      {
         try
         {
            replacement = (String) pendingStack.pop();
         }
         catch (EmptyStackException exc)
         {
            // TODO: proper logging
            exc.printStackTrace();
         }
      }
      
      return replacement;
   }
   
   /**
    * Perform preparation of the data structures used to match candidate
    * substrings in the {@link #nextToken} method. This includes setting
    * each candidate key string into an array, and creating a new array
    * of integers to track the match status of each candidate key as the
    * <code>nextToken</code> method scans characters from the original
    * symbol.
    */
   private void prepareCandidates()
   {
      if (candidates != null)
      {
         return;
      }
      
      Set keys = replacements.keySet();
      int size = keys.size();
      candidates = new String[size];
      matches = new int[size];
      
      Iterator iter = keys.iterator();
      for (int i = 0; iter.hasNext(); i++)
      {
         String next = (String) iter.next();
         candidates[i] = next;
      }
   }
   
   /**
    * Test the given name against the list of Java keywords and resolve any
    * conflict by adding text to the name to make it different (and thus
    * javac will compile it).  An underscore character will be prepended
    * to the name to resolve any conflicts.
    *
    * @param    possible
    *           The name to check for conflicts.
    * @param    type
    *           The type of name.  This is used to ensure that the name is
    *           a Java name (and thus requires conflict resolution).
    *
    * @return   The original name if no conflicts were detected or a
    *           disambiguated (safe) name if a conflict existed.
    */
   private String resolvePossibleKeywordConflict(String possible, int type)
   {
      if (type == TYPE_VARIABLE ||
          type == TYPE_METHOD   ||
          type == TYPE_CONSTANT ||
          type == TYPE_CLASS)
      {
         if (reserved.contains(possible))
         {
            StringBuilder sb = new StringBuilder(possible);
            possible = sb.append("_").toString();
         }
      }

      return possible;
   }
   
   /**
    * Entry point and test harness for this class. Expects the following
    * arguments:
    * <ol>
    * <li><code>symbol</code> (required) - Progress symbol name to be
    *     converted;
    * <li><code>type</code> (required) - type of conversion to apply. Use
    *     the integer equivalents of the <code>TYPE_*</code> constants
    *     defined by this class;
    * <li><code>replace</code> (required) - true/false. True to replace
    *     substrings as the name is tokenized; false to tokenize syllables
    *     without replacement;
    * <li><code>key</code> (optional) - substring in <code>symbol</code>
    *     to be replaced by the following parameter;
    * <li><code>replacement</code> (required) - replacement text for the
    *     immediately previous parameter.
    * </ol>
    * Any number of key/replacement pairs may be provided. If a key is
    * provided with no following replacement, that key is ignored.
    *
    * @param   args
    *          Command line parameters as described above.
    */
   public static void main(String[] args)
   {
      try
      {
         if (args.length < 3)
         {
            System.out.println(
               "Usage: NameConverter \"symbol\" type replace phrases [\"key\", \"replacement\"]...");
            System.exit(-1);
         }
         
         String name = args[0];
         int type = Integer.parseInt(args[1]);
         boolean replace = (new Boolean(args[2])).booleanValue();
         int phrases = ALL;
         
         if (args.length > 3)
         {
            phrases = Integer.parseInt(args[3]);
         }
         NameConverter conv = new NameConverter(phrases, replace);
         
         int i;
         for (i = 4; i + 1 < args.length; i += 2)
         {
            String key = args[i];
            String rpl = args[i + 1];
            conv.putReplacement(key, rpl);
         }
         
         if (i + 1 == args.length)
         {
            System.out.println("Extra argument '" + args[i] + "' ignored");
         }
         
         System.out.println(name + " ---> " + conv.convert(name, type) +
                            " (" + conv.describeConversion() + ")");
         /*
         long time = System.currentTimeMillis();
         for (i = 0; i < 10000; i++)
         {
            conv.convert(name, type);
         }
         time = System.currentTimeMillis() - time;
         System.out.println("Elapsed:  " + time);
         */
      }
      catch (Exception exc)
      {
         exc.printStackTrace();
      }
   }
}
